@@ -24,14 +24,17 @@ from email.errors import HeaderParseError
 from email.header import decode_header
 from unicodedata import normalize
 
+import base64
 import datetime
 import email
 import functools
 import hashlib
 import logging
 import os
+import random
 import re
 import simplejson as json
+import string
 import subprocess
 import sys
 import tempfile
@@ -44,7 +47,7 @@ from .const import (
     OTHERS_PARTS,
     RECEIVED_COMPILED_LIST)
 
-from .exceptions import MailParserOSError
+from .exceptions import MailParserOSError, MailParserReceivedParsingError
 
 
 log = logging.getLogger(__name__)
@@ -240,6 +243,58 @@ def msgconvert(email):
         os.close(temph)
 
 
+def parse_received(received):
+    """
+    Parse a single received header.
+    Return a dictionary of values by clause.
+
+    Arguments:
+        received {str} -- single received header
+
+    Raises:
+        MailParserReceivedParsingError -- Raised when a
+            received header cannot be parsed
+
+    Returns:
+        dict -- values by clause
+    """
+
+    values_by_clause = {}
+    for pattern in RECEIVED_COMPILED_LIST:
+        matches = [match for match in pattern.finditer(received)]
+
+        if len(matches) == 0:
+            # no matches for this clause, but it's ok! keep going!
+            log.debug("No matches found for %s in %s" % (
+                pattern.pattern, received))
+            continue
+        elif len(matches) > 1:
+            # uh, can't have more than one of each clause in a received.
+            # so either there's more than one or the current regex is wrong
+            msg = "More than one match found for %s in %s" % (
+                pattern.pattern, received)
+            log.error(msg)
+            raise MailParserReceivedParsingError(msg)
+        else:
+            # otherwise we have one matching clause!
+            log.debug("Found one match for %s in %s" % (
+                pattern.pattern, received))
+            match = matches[0].groupdict()
+            if six.PY2:
+                values_by_clause[match.keys()[0]] = match.values()[0]
+            elif six.PY3:
+                key = list(match.keys())[0]
+                value = list(match.values())[0]
+                values_by_clause[key] = value
+
+    if len(values_by_clause) == 0:
+        # we weren't able to match anything...
+        msg = "Unable to match any clauses in %s" % (received)
+        log.error(msg)
+        raise MailParserReceivedParsingError(msg)
+    return values_by_clause
+
+
 def receiveds_parsing(receiveds):
     """
     This function parses the receiveds headers.
@@ -256,31 +311,31 @@ def receiveds_parsing(receiveds):
     n = len(receiveds)
     log.debug("Nr. of receiveds. {}".format(n))
 
-    try:
-        # Loop receiveds
-        for j, i in enumerate(receiveds):
-            log.debug("Parsing received {}/{}".format(j + 1, n))
-            log.debug("Try to parse {!r}".format(i))
-            # Try more patterns
-            for p in RECEIVED_COMPILED_LIST:
-                log.debug("Try to parse with {!r}".format(p.pattern))
-                # Test if all string match
-                t = p.search(i)
-                if t and t.group() == i:
-                    log.debug("Received parsed {!r}".format(i))
-                    parsed.append(t.groupdict())
-                    break
-                else:
-                    log.debug("Received not parsed {!r}".format(i))
-                    continue
+    for idx, received in enumerate(receiveds):
+        log.debug("Parsing received {}/{}".format(idx + 1, n))
+        log.debug("Try to parse {!r}".format(received))
+        try:
+            # try to parse the current received header...
+            values_by_clause = parse_received(received)
+        except MailParserReceivedParsingError:
+            # if we can't, let's append the raw
+            parsed.append({'raw': received})
         else:
-            if len(receiveds) != len(parsed):
-                raise ValueError
+            # otherwise append the full values_by_clause dict
+            parsed.append(values_by_clause)
 
-    except (AttributeError, ValueError):
+    log.debug("len(receiveds) %s, len(parsed) %s" % (
+        len(receiveds), len(parsed)))
+
+    if len(receiveds) != len(parsed):
+        # something really bad happened,
+        # so just return raw receiveds with hop indices
+        log.error("len(receiveds): %s, len(parsed): %s, receiveds: %s, \
+            parsed: %s" % (len(receiveds), len(parsed), receiveds, parsed))
         return receiveds_not_parsed(receiveds)
 
     else:
+        # all's good! we have parsed or raw receiveds for each received header
         return receiveds_format(parsed)
 
 
@@ -291,8 +346,9 @@ def convert_mail_date(date):
     t = email.utils.mktime_tz(d)
     log.debug("Date parsed in timestamp: {!r}".format(t))
     date_utc = datetime.datetime.utcfromtimestamp(t)
-    timezone = d[9] / 3600 if d[9] else 0
-    timezone = "{:+.0f}".format(timezone)
+    timezone = d[9] / 3600.0 if d[9] else 0
+    timezone = "{:+.1f}".format(timezone)
+    log.debug("Calculated timezone: {!r}".format(timezone))
     return date_utc, timezone
 
 
@@ -402,27 +458,42 @@ def get_header(message, name):
         name (string): header to get
 
     Returns:
-        decoded header
+        str if there is an header
+        list if there are more than one
     """
-    header = message.get(name)
-    log.debug("Getting header {!r}: {!r}".format(name, header))
-    if header:
-        return decode_header_part(header)
+
+    headers = message.get_all(name)
+    log.debug("Getting header {!r}: {!r}".format(name, headers))
+    if headers:
+        headers = [decode_header_part(i) for i in headers]
+        if len(headers) == 1:
+            # in this case return a string
+            return headers[0]
+        # in this case return a list
+        return headers
     return six.text_type()
 
 
-def get_mail_keys(message):
+def get_mail_keys(message, complete=True):
     """
     Given an email.message.Message, return a set with all email parts to get
 
     Args:
         message (email.message.Message): email message object
+        complete (bool): if True returns all email headers
 
     Returns:
         set with all email parts
     """
-    all_headers_keys = {i.lower() for i in message.keys()}
-    all_parts = ADDRESSES_HEADERS | OTHERS_PARTS | all_headers_keys
+
+    if complete:
+        log.debug("Get all headers")
+        all_headers_keys = {i.lower() for i in message.keys()}
+        all_parts = ADDRESSES_HEADERS | OTHERS_PARTS | all_headers_keys
+    else:
+        log.debug("Get only mains headers")
+        all_parts = ADDRESSES_HEADERS | OTHERS_PARTS
+
     log.debug("All parts to get: {}".format(", ".join(all_parts)))
     return all_parts
 
@@ -446,7 +517,7 @@ def print_attachments(attachments, flag_hash):  # pragma: no cover
     if flag_hash:
         for i in attachments:
             if i.get("content_transfer_encoding") == "base64":
-                payload = i["payload"].decode("base64")
+                payload = base64.b64decode(i["payload"])
             else:
                 payload = i["payload"]
 
@@ -455,3 +526,50 @@ def print_attachments(attachments, flag_hash):  # pragma: no cover
 
     for i in attachments:
         safe_print(json.dumps(i, ensure_ascii=False, indent=4))
+
+
+def write_attachments(attachments, base_path):  # pragma: no cover
+    for a in attachments:
+        write_sample(
+            binary=a["binary"],
+            payload=a["payload"],
+            path=base_path,
+            filename=a["filename"],
+        )
+
+
+def write_sample(binary, payload, path, filename):  # pragma: no cover
+    """
+    This function writes a sample on file system.
+
+    Args:
+        binary (bool): True if it's a binary file
+        payload: payload of sample, in base64 if it's a binary
+        path (string): path of file
+        filename (string): name of file
+        hash_ (string): file hash
+    """
+    if not os.path.exists(path):
+        os.makedirs(path)
+
+    sample = os.path.join(path, filename)
+
+    if binary:
+        with open(sample, "wb") as f:
+            f.write(base64.b64decode(payload))
+    else:
+        with open(sample, "w") as f:
+            f.write(payload)
+
+
+def random_string(string_length=10):
+    """ Generate a random string of fixed length
+
+    Keyword Arguments:
+        string_length {int} -- String length (default: {10})
+
+    Returns:
+        str -- Random string
+    """
+    letters = string.ascii_lowercase
+    return ''.join(random.choice(letters) for i in range(string_length))
